// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{kw, module::transform_body::Visitor, HdlAttr};
use quote::{format_ident, quote_spanned};
use syn::{
    parse::Nothing, parse_quote, parse_quote_spanned, spanned::Spanned, Expr, ExprArray, ExprPath,
    ExprRepeat, ExprStruct, ExprTuple, FieldValue, TypePath,
};

impl Visitor<'_> {
    pub(crate) fn process_hdl_array(
        &mut self,
        hdl_attr: HdlAttr<Nothing, kw::hdl>,
        mut expr_array: ExprArray,
    ) -> Expr {
        self.require_normal_module_or_fn(hdl_attr);
        for elem in &mut expr_array.elems {
            *elem = parse_quote_spanned! {elem.span()=>
                ::fayalite::expr::ToExpr::to_expr(&(#elem))
            };
        }
        parse_quote! {::fayalite::expr::ToExpr::to_expr(&#expr_array)}
    }
    pub(crate) fn process_hdl_repeat(
        &mut self,
        hdl_attr: HdlAttr<Nothing, kw::hdl>,
        mut expr_repeat: ExprRepeat,
    ) -> Expr {
        self.require_normal_module_or_fn(hdl_attr);
        let repeated_value = &expr_repeat.expr;
        *expr_repeat.expr = parse_quote_spanned! {repeated_value.span()=>
            ::fayalite::expr::ToExpr::to_expr(&(#repeated_value))
        };
        parse_quote! {::fayalite::expr::ToExpr::to_expr(&#expr_repeat)}
    }
    pub(crate) fn process_hdl_struct(
        &mut self,
        hdl_attr: HdlAttr<Nothing, kw::hdl>,
        expr_struct: ExprStruct,
    ) -> Expr {
        self.require_normal_module_or_fn(&hdl_attr);
        let name_span = expr_struct.path.segments.last().unwrap().ident.span();
        let builder_ident = format_ident!("__builder", span = name_span);
        let empty_builder = if expr_struct.qself.is_some()
            || expr_struct
                .path
                .segments
                .iter()
                .any(|seg| !seg.arguments.is_none())
        {
            let ty = TypePath {
                qself: expr_struct.qself,
                path: expr_struct.path,
            };
            let builder_ty = quote_spanned! {name_span=>
                <#ty as ::fayalite::bundle::BundleType>::Builder
            };
            quote_spanned! {name_span=>
                <#builder_ty as ::fayalite::__std::default::Default>::default()
            }
        } else {
            let path = ExprPath {
                attrs: vec![],
                qself: expr_struct.qself,
                path: expr_struct.path,
            };
            quote_spanned! {name_span=>
                #path::__bundle_builder()
            }
        };
        let field_calls = Vec::from_iter(expr_struct.fields.iter().map(
            |FieldValue {
                 attrs: _,
                 member,
                 colon_token: _,
                 expr,
             }| {
                let field_fn = format_ident!("field_{}", member);
                quote_spanned! {member.span()=>
                    let #builder_ident = #builder_ident.#field_fn(#expr);
                }
            },
        ));
        parse_quote_spanned! {name_span=>
            {
                let #builder_ident = #empty_builder;
                #(#field_calls)*
                ::fayalite::expr::ToExpr::to_expr(&#builder_ident)
            }
        }
    }
    pub(crate) fn process_hdl_tuple(
        &mut self,
        hdl_attr: HdlAttr<Nothing, kw::hdl>,
        expr_tuple: ExprTuple,
    ) -> Expr {
        self.require_normal_module_or_fn(hdl_attr);
        parse_quote_spanned! {expr_tuple.span()=>
            ::fayalite::expr::ToExpr::to_expr(&#expr_tuple)
        }
    }
}
